function bertin_multipitch(wavfile,framewise_f0file,notewise_f0file)
% Fast Bayesian constrained Itakura-Saito NMF applied to a ERBT, with
% basis spectra representing partial clusters and fundamental frequencies
% on the MIDI scale tuned at 440 Hz
% Framewise f0 transcription every 10 ms / Transcription of the onset,
% offset and f0 of each note
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

display(sprintf('processing %s',wavfile));

% Reading WAV file
[x,fs]=wavread(wavfile);
x=resample(x,22050,fs).';
fs=22050;

% Defining optimal parameters and constants
nbfreq=250; maxclus=8; cluswidth=3;
wlen = 2^nextpow2(.02*fs);
[I,T]=size(x);
N=ceil(T/wlen);
K=88;
nb_iter=100;
seuil=-55;
eta = 0.5;

% Computing ERBT coefficients and frequency scale
display('computing ERBT coefficients and frequency scale');
V=zeros(nbfreq,N,I);
for i=1:I,
   [V(:,:,i),f]=erbtm(x(i,:),fs,nbfreq,wlen);
end
V=sum(V.^2,3)+1e-9;

% Applying NMF
display('applying NMF');
[U,A] = phsnmfmu(V,f,fs,nb_iter,K,maxclus,cluswidth,eta);


display('write outputs');

% Estimating onset/offset
testnotes=harm2midi_erbt(U(:,1:88),A(1:88,:),fs,seuil);

% Writes notewise file
f0file = notewise_f0file;
nbnotes=length(testnotes);
fid=fopen(f0file,'w');
for n=1:nbnotes,
    fprintf(fid,'0 0 %i 0 %.3f %.3f\n',testnotes(n,3),testnotes(n,5),testnotes(n,6));
end
fclose(fid);

% converting to f0
testnotes(:,3)=440*2.^((testnotes(:,3)-69)/12);

% Transcribing into framewise format
N=ceil(length(x)/220.5);
test=zeros(N,1);
for note=1:size(testnotes,1),
    b=max(1,floor(testnotes(note,5)*100)+1);
    e=min(N,floor(testnotes(note,6)*100)+1);
    p=1;
    while any(test(b:e,p)),
        p=p+1;
        if p > size(test,2),
            test=[test zeros(N,1)];
        end
    end
    test(b:e,p)=testnotes(note,3);
end
test=test(1:N,:);

% Writing into a file
f0file = framewise_f0file;
NbColumns = 30;
fid=fopen(f0file,'w');
for n=1:N,
    fprintf(fid,'%.2f',(n-1)/100);
    testloc=test(n,test(n,:)~=0);
    nbloc=length(testloc);
    if (nbloc>=NbColumns-1),
        nbloc = NbColumns-1;
    end
    for p=1:nbloc,
        fprintf(fid,'\t%.2f',testloc(p));
    end
    if nbloc<NbColumns,
        for p=nbloc+1:NbColumns-1,
            fprintf(fid,'\t0.00');
        end
    end
    fprintf(fid,'\n');
end

fclose(fid);

function [X,f]=erbtm(x,fs,F,wlen)

% ERBTM Magnitude ERB Transform using a Hann window.
%
% [X,f]=erbtp(x,fs,F,wlen)
%
% Inputs:
% x: 1 x T vector containing a single-channel signal
% fs: sampling frequency in Hz
% F: number of frequency bins (the ratio between the bandwidth of each bin
% and the frequency difference between successive bins is constant)
% wlen: number of samples per frame (must be a multiple of the largest
% downsampling factor, typically a large power of 2)
%
% Output:
% X: F x N matrix containing the time-frequency magnitude (amplitude) coefficients
% f: F x 1 vector containing the center frequency of each frequency bin

%%% Errors and warnings %%%
if nargin<4, error('Not enough input arguments.'); end
[I,T]=size(x);
if I>1, error('The input signal must be a row vector.'); end
N=ceil(T/wlen);

%%% Computing ERBT coefficients %%%
x=hilbert(x);
X=zeros(F,N);
% Determining minimum and maximum frequency
fmax=.5*fs; fmin=0;
for j=1:100,
    emin=9.26*log(.00437*fmin+1);
    emax=9.26*log(.00437*fmax+1);
    fmin=1.5*(emax-emin)/(F-1)/9.26/.00437*exp(emin/9.26);
    fmax=.5*fs-1.5*(emax-emin)/(F-1)/9.26/.00437*exp(emax/9.26);
    if (fmax < 0) || (fmin > .5*fs), error('The number of frequency bins is too small.'); end
end
% Determining frequency and window length scales
emin=9.26*log(.00437*fmin+1);
emax=9.26*log(.00437*fmax+1);
e=(0:F-1)*(emax-emin)/(F-1)+emin;
f=(exp(e/9.26)-1)/.00437;
a=.5*(F-1)/(emax-emin)*9.26*.00437*fs*exp(-e/9.26)-.5;
% Determining dyadic downsampling bins (for fast computation)
fup=f+1.5*fs./(2*a+1);
subs=-log(2*fup/fs)/log(2);
subs=2.^max(0,floor(min(log2(wlen),subs)));
if (wlen/subs(1) ~= floor(wlen/subs(1))), error(['The number of samples per frame must be a multiple of ' int2str(subs(1)) '.']); end
down=(subs~=[subs(2:end),1]);
for bin=F:-1:1,
    % Dyadic downsampling
    if down(bin),
        x=resample(x,1,2,50);
    end
    % Convolution with a modulated sine window
    hwlen=round(a(bin)/subs(bin));
    win=hanning(2*hwlen+1).'.*exp(1i*2*pi*f(bin)/fs*subs(bin)*(-hwlen:hwlen));
    band=[fftfilt(win,[x,zeros(1,2*hwlen)]) zeros(1,wlen/subs(bin))];
    % Square-root energy on short time frames
    band=band(hwlen+1:hwlen+N*wlen/subs(bin));
    X(bin,:)=sqrt(sum(reshape(abs(band).^2,wlen/subs(bin),N),1)/(hwlen+1)^2*subs(bin));
end

return;

function [U,A] = phsnmfmu(PowSpec,f,fs,nb_iter_mu,K,maxclus,cluswidth,eta);

% NMF minimizing Itakura-Saito distance, multiplicative version (Roland)
% Harmonicity model on W
% Smoothness model (Markov Inverse-Gamma) on H
% Inputs :
%   - PowSpec: positive matrix data (F x N)
%   - f: frequency centers of ERB bins
%   - n_iter_mu: number of iterations for multiplicative NMF phase


% parameters % debug: ? passer en entr?e
[F,N]=size(PowSpec);

% initializations
h = abs(1+randn(K,N));
alpha_h = 10*ones(K,1);

Kf=32;

%%% HARMONIC STUFF %%%

nbcomp=1;
fmin=f(1); fmax=f(F);
emin=9.26*log(.00437*fmin+1); emax=9.26*log(.00437*fmax+1);
e=(0:F-1)*(emax-emin)/(F-1)+emin;
a=.5*(F-1)/(emax-emin)*9.26*.00437*fs*exp(-e/9.26)-.5;
alen=2*round(a)+1;
f=f/fs;

%%% Defining partial spectra %%%
firstnote=21; lastnote=108; nbnotes=lastnote-firstnote+1; pitch=firstnote:lastnote;
f0=2.^((pitch-69)/12)*440/fs;
nharm=floor(.5./f0);
ppos=[0,cumsum(nharm)];
nbpart=ppos(end);
Z=zeros(F,nbpart);   %partial spectra
partfreq=zeros(1,nbpart);
for n=1:nbnotes,
    partfreq(ppos(n)+1:ppos(n)+nharm(n))=f0(n)*(1:nharm(n));
end
for c=1:F,
    Z(c,:)=abs(sinc((f(c)-partfreq)*alen(c))+.5*sinc((f(c)-partfreq)*alen(c)+1)+.5*sinc((f(c)-partfreq)*alen(c)-1));
end

%%% Defining cluster spectra and initializing note spectra %%%
e0=9.26*log(.00437*f0*fs+1);
parterb=9.26*log(.00437*partfreq*fs+1);
nclus=min(maxclus,floor((emax-e0)/cluswidth)+1);
cpos=[0,cumsum(nclus)];
nbclus=cpos(end);
V=zeros(F,nbclus);   %cluster spectra
B=zeros(nbclus,nbcomp);  %cluster weights
U=zeros(F,nbnotes*nbcomp);  %note spectra
for n=1:nbnotes,
    weights=zeros(nharm(n),nclus(n));
    for c=1:nclus(n),
        weights(:,c)=(1+((parterb(ppos(n)+1:ppos(n)+nharm(n)).'-e0(n)-cluswidth*(c-1))/(1.019*cluswidth)).^2).^-2;
    end
    V(:,cpos(n)+1:cpos(n)+nclus(n))=Z(:,ppos(n)+1:ppos(n)+nharm(n))*weights;
    %modif Emmanuel meilleure pente
    eslope = 1;
    B(cpos(n)+1:cpos(n)+nclus(n),:)=(10.^(-eslope*cluswidth/20*(0:nclus(n)-1)).'*ones(1,nbcomp)).^(ones(nclus(n),1)*(1./(1:nbcomp)));
    U(:,(0:nbcomp-1)*nbnotes+n)=V(:,cpos(n)+1:cpos(n)+nclus(n))*B(cpos(n)+1:cpos(n)+nclus(n),:);
end
% %%% END OF HARMONIC STUFF %%%

scale = sqrt(sum(U.^2));
for k=1:K
    B(cpos(k)+1:cpos(k)+nclus(k)) = B(cpos(k)+1:cpos(k)+nclus(k))/scale(k);
    h(k,:) = h(k,:)*scale(k);
    U(:,k)=V(:,cpos(k)+1:cpos(k)+nclus(k))*B(cpos(k)+1:cpos(k)+nclus(k));
end

PSDf = abs(1+randn(F,Kf));
Af = abs(1+randn(Kf,N));
A=h;
X=PowSpec;
PowSpec_app = U*h;
PSDtot = [U,PSDf];
Atot = [A;Af];
PowSpectot = PSDtot*Atot;

for kk=2:nb_iter_mu+1

    % H updates

    Pm2 = PowSpec_app.^-2;   
    
    num1 = alpha_h(1)./A(:,1);
    numn = (alpha_h(1)+1)*A(:,1:end-1)./A(:,2:end).^2;
    num = [num1,numn];
    
    denomn = 1./A(:,1:end-1) + (alpha_h(1)+1)./A(:,2:end);
    denomN = (alpha_h(1)+1)./A(:,N);
    denom = [denomn,denomN];

    A = A .* ((U'*(X.*Pm2)+ num)./(U'*(PowSpec_app.^-1) + denom)).^eta;
    
    % maj des composantes libres
    Af = Af .* (PSDf.'*(X.*(PowSpectot.^(-2))))./(PSDf.'*(PowSpectot.^(-1)));
    Atot = [A;Af];   
    PSDtot = [U,PSDf];
    PowSpec_app = PSDtot*Atot;

    XP=X.*PowSpec_app.^-2;
    Pm1=PowSpec_app.^-1;
    for k=1:nbnotes,

%        for m=1:nclus(k)
%           hkn_Pkmf = (ones(F,1)*A(k,:)).*(V(:,cpos(k)+m)*ones(1,N));
%           B(cpos(k)+m) = B(cpos(k)+m)*((sum(sum(hkn_Pkmf.*XP)) ./ sum(sum(hkn_Pkmf.*Pm1)))).^eta;
%        end
        
        % 16/09/2009, BM: La boucle precedente peut etre avantageusement
        % remplacee par les 2 lignes suivantes pour un calcul identique.
        m = 1:nclus(k);
        B(cpos(k)+m) = B(cpos(k)+m) .* (((A(k,:)*XP')*V(:,cpos(k)+m)) ./ ((A(k,:)*Pm1')*V(:,cpos(k)+m)))'.^eta;
        
        U(:,(0:nbcomp-1)*nbnotes+k)=V(:,cpos(k)+1:cpos(k)+nclus(k))*B(cpos(k)+1:cpos(k)+nclus(k),:);
    end

    PSDf = PSDf .* ((X.*PowSpec_app.^-2)*Af')./(PowSpec_app.^-1*Af');
    
    for k=1:K
        U(:,k)=V(:,cpos(k)+1:cpos(k)+nclus(k))*B(cpos(k)+1:cpos(k)+nclus(k));
        scale = sqrt(sum(U(:,k).^2));
        B(cpos(k)+1:cpos(k)+nclus(k)) = B(cpos(k)+1:cpos(k)+nclus(k))/scale;
        A(k,:) = A(k,:)*scale;
        U(:,k)=V(:,cpos(k)+1:cpos(k)+nclus(k))*B(cpos(k)+1:cpos(k)+nclus(k));
    end
    
    Atot = [A;Af];   
    PSDtot = [U,PSDf];
    PowSpec_app = PSDtot*Atot;
    
    for k=1:12
        scale = sqrt(sum(PSDf(:,k).^2));
        PSDf(:,k) = PSDf(:,k)./scale;
        Af(k,:) = Af(k,:)*scale;
    end  
    
end
A = [A;Af];
U = [U,PSDf];

A=A.*(sqrt(sum(U.^2)).'*ones(1,N));
U=U./(ones(F,1)*sqrt(sum(U.^2)));


function nmat=harm2midi_erbt(W,H,fs,seuil)

% Outputs :
%   nmat : midi-like matrix of the transcription result (Ken Schutte's midi
%   toolbox format)
%   note minimal duration equal to 50 ms whatever the window length (30 ms
%   in practice for the chosen test files)
%
% Inputs :
%   W : dictionary of atoms (first NMF output)
%   H : temporal activities matrix (second NMF output)
%   fs : sampling rate

n=size(H,2);
r=size(W,2);
wlen=2^nextpow2(.02*fs);

% Pitch definition
firstnote=21; lastnote=108; nbnotes=lastnote-firstnote+1;
nbcomp=r/nbnotes;
pitch=reshape((firstnote:lastnote).'*ones(1,nbcomp),1,nbnotes*nbcomp);

% Summation of components with the same pitch
k=1;
while k < size(H,1)
    currpitch = pitch(k);
    indices = find(pitch(k+1:end)==currpitch);
    if ~isempty(indices)
        X=W(:,[k indices+k])*H([k indices+k],:);
        H(k,:) = sqrt(sum(X.^2));
        H(indices+k,:)=[];
        pitch(indices+k)=[];
    end
    k = k+1;
end
r = size(H,1);

% Minimal note magnitude and duration
thresh=max(max(H))*10^(seuil/20);
mindur=ceil(.05*fs/wlen);

% TH coefficients are 1 for a "note ON" event
TH = zeros(r,n);
TH(:,1) = all((H(:,1:mindur)>thresh),2);
for k=2:n-mindur+1
    TH(:,k) = (H(:,k-1)<=thresh) & all((H(:,k:k+mindur-1)>thresh),2);
end

% DH coefficients are 1 for a "note OFF" event
DH = zeros(r,n);
for k=mindur:n-1
    DH(:,k) = (H(:,k+1)<=thresh) & all((H(:,k-mindur+1:k)>thresh),2);
end
DH(:,n) = all((H(:,n-mindur+1:n)>thresh),2);

% initialize notematrix
pos_events = find(TH==1);
nb_events = length(pos_events);
nmat = zeros(nb_events,8);

% builds output matrix
for p=1:nb_events
    [atom,timeon]=ind2sub([r n],pos_events(p));
    timeoff=find(DH(atom,timeon+1:end)); timeoff=timeon+timeoff(1);
    nmat(p,3)=pitch(atom);
    nmat(p,4)=100;
    nmat(p,5)=(timeon-.5)*wlen/fs;
    nmat(p,6)=(timeoff-.5)*wlen/fs;
end
nmat(:,1)=3;
nmat(:,2)=0;
nmat(:,7)=2*cumsum(ones(size(nmat,1),1))-1;
nmat(:,8)=2*cumsum(ones(size(nmat,1),1));

% delete unrelevant events
nmat=nmat((nmat(:,3)>0),:);

return;


















