function dst = enhanced_resample (src, src_param, dst_param)
% dst = enhanced_resample (src, src_param, dst_param)
%  where param is
% - either the sampling rate as a scalar (uniform case)
% - or the sampling times in a vector (nonunifom case)
%
% resample src signal according to the specified rates or times

% TODO: Burg extrapolation
% TODO: nonuniform src
% TODO: explain why no optimization for the uniform / uniform rational case
%  (cannot avoid loop for convolution -- because of local polynomial approximation,
%   but could avoid recomputing the reconstructor at each iteration...)
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

  src_uniform = isscalar (src_param);
  dst_uniform = isscalar (dst_param);

  k = 10;  % number of wings for each side
  order = 3;  % polynomial order
  tol = eps*1e3;  % epsilon for float comparison / rounding

  % get rates
  if src_uniform
    % uniform case
    src_rate = src_param;
  else
    % general case
    assert (length(src) == length(src_param));
    [src_time permut] = sort (src_param);  % sort src times in chronological order...
    src = src(permut);  % ...then order the src samples accordingly
    src_rate = 1 / median(diff(dst_time));  %#ok<NODEF> % should be mode... better than mean?
  end
  if dst_uniform
    % uniform case
    dst_rate = dst_param;
  else
    % general case
    [dst_time permut] = sort (dst_param);  % sort dst times in chronological order...
    dst_order(permut) = (1:length(dst_param));  % ...and find the way back to the original order
    dst_rate = 1 / median(diff(dst_time));  % should be mode... better than mean?
  end

  rate = min(src_rate,dst_rate);
  ratio = rate/src_rate;
  T_half = k*ratio/src_rate;

  if src_uniform
    % uniform case:
    % fix boundaries (reflection method)
    left = ceil(T_half*src_rate-tol);
    before = zeros(1,left);
    before(left:-1:1) = 2*src(1) - src(1+(1:left));
    src_size = length(src);
    right = floor(T_half*src_rate+tol);
    after = zeros(1,right);
    after(1:right) = 2*src(src_size) - src(src_size-(1:right));
    all = [before src after];
    prefix = length(before);
    src_time = ((0:length(all)-1)-prefix) / src_rate;
  else
    % general case:
    % cannot do anything?
    all = src;
    prefix = 0;
    src_time = src_param;
    assert (src_size == length(src_time)); %#ok<NODEF>
  end

  if dst_uniform
    % uniform case
    dst_size = ceil(dst_rate*src_size/src_rate-tol);
    dst = zeros (1, dst_size);
    dst_time = (0:dst_size-1) / dst_rate;
  else
    % general case
    dst_time = dst_param;
    dst_size = length(dst_time);
  end

% resampling loop
  for n=1:dst_size
    t_dst = dst_time(n);
    if src_uniform
      % small optimization...
      left = ceil((t_dst-T_half)*src_rate-tol);
      right = floor((t_dst+T_half)*src_rate+tol);
      span = (left:right);
      t_src = span / src_rate;
    else
      span = find(abs(src_time-t_dst)<=T_half+tol)-1;  % bloody MatLab indices...
      t_src = src_time(span);
    end
    delta_t = t_dst - t_src;
    rw = 0.5*(1-cos(pi*delta_t/T_half-pi));  % (Hann) window (zero-phase)
%   rw = 0.35875 - 0.48829 * cos(pi*delta_t/T_half-pi) + 0.14128 * cos(2*pi*delta_t/T_half) - 0.01168 *cos(3*pi*delta_t/T_half-pi);  % (Blackman-Harris) window (zero-phase)
    ri = sinc(delta_t/T_half*k);  % ideal reconstructor (truncated)
    r = rw .* ri;  % practical reconstructor
    r = r / sum(r);  % normalized
    % for the reminder, see Martin Raspaud and Sylvain Marchand,
    % "Enhanced Resampling for Sinusoidal Modeling Parameters", in Proc. IEEE WASPAA 2007
    s = all(span+prefix+1);
    t_min = min(t_src);  % time offset
    x = t_src - t_min;  % remove time offset to reduce instability problems with polynomials
    [p, mu] = superpolyfit(x,s,order);  % estimate polynomial on actual reconstructor span
%   p = 0;
    s = s - superpolyval(p,mu,x);  % and subtract is from the signal segment
    dst(n) = sum (s .* r) + superpolyval(p,mu,t_dst-t_min);  % convolution evaluated at sample n + polynomial evaluated at desired time
  end

  if not(dst_uniform)
    dst = dst(dst_order);
  end
end
