function [freqs,amps] = hrhatrack(s,Ns,P,beta,muL,muV)

% [freqs,amps] = hrhatrack(s,Ns,P,beta,muL,muV)
% freqs : estimated frequency trajectories (normalized frequency, Nyquist = 0.5)
% amps  : corresponding instantaneous amplitudes 
% s : data, can be complex
% Ns : size of the signal subspace. Ideally, the number of complex 
%      frequency components, if it is known. If not known, to
%      overestimated. ex : for 2 real modulated sinusoids, take Ns = 4. If
%      it is assumed to be at least 5, try between 14 and 20. 
%      Default :  Ns = 8;
% P : the autocovariance matrix is of size PxP, default P = 3*Ns;
% beta : forgetting factor for the estimated covariance matrix
%        (Rss(t+1) = beta Rss(t) + x'x
%        default: 0.99;
% muL : gradient step for the eigenvalues, default 0.9
% muV : gradient step for the eigenvectors, default 0.9

% refs: 
% [1] Fast Approximated Power Iteration Subspace Tracking, Roland Badeau,
% Bertrand David et Gal Richard, IEEE Transactions on Signal Processing, 
% vol. 53, no. 8, pp. 2931-2941, aot 2005
% [2] HRHATRAC Algorithm for Spectral Line Tracking of Musical Signals,
% Bertrand David, Roland Badeau et Gal Richard, 
% 2006 International Conference on Acoustics, Speech, 
% and Signal Processing ICASSP06, Toulouse, France, 15-19 mai 2006, 
% vol. III, pp. 45-48
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

if nargin < 6, muV = 0.9; end
if nargin < 5, muL = 0.9; end
if nargin < 4, beta = 0.99; end
if nargin < 2,Ns = 8;end
if nargin < 3, P = 3*Ns;end


%%%%%%%%%%%%%%%%%%%%%
% initialization
%%%%%%%%%%%%%%%%%%%%%

Nh = 2*P; % analysis length
s = s(:,1);N = length(s);
W = orth(randn(P,Ns)); % initialization of the signal subspace to a random
                       % matrix
Z = eye(Ns);
Wdown = W(1:end-1,:);
Wup = W(2:end,:);
Phi =pinv(Wdown)*Wup; % spectral matrix
[G,J]=eig(Phi);
Psi = Wdown'*Wup;

% FAPI init [1]
er = zeros(1,N-P);
freqs = zeros(Ns,N-P);
amps = zeros(Ns,N-P);

ort = zeros(P-1,N-P);
Kt = zeros(1,N-P);
n=0:P-1;
Nit = N-P;


hw =waitbar(0,'calcul');

for k= 1:Nit
    %%%%%%%%%%%%%%%%%%%%%%%%
    % FAPI [1] section
    %%%%%%%%%%%%%%%%%%%%%%%%
    x = s(k+1:k+P);
    
    y = W'*x;
    h = Z*y;
    g = h / (beta + y'*h);
    
    eps2 = x'*x - y'*y;
    ng = g'*g;
    tau = eps2 / (1 + eps2*ng + sqrt(1+eps2*ng) );
    eta = 1 - tau*ng;
    yp = eta*y + tau*g;
    hp = Z'*yp;
    epsi = tau/eta*(Z*g- (hp'*g)*g);
    Z = 1/beta*(Z - g*hp' + epsi*g');
    ep = eta*x - W*yp;
    
   
    edown = ep(1:end-1);eup = ep(2:end);
    emu = W(1:end-1,:)'*eup;
    epu = W(2:end,:)'*edown;
    nutmun = W(end,:)';
    
    W = W + ep*g'; % update of the signal subspace
    % end FAPI
    
    % update of the spectral matrix
    nut = W(end,:)';
    epup = epu + g*(eup'*edown);
    phitmun = Psi'*nutmun;
    Psi = Psi + emu*g'+g*epup';
    phit = Psi'*nut;
    at = 1/(1-real(nut'*nut)) * nut ;
    Phiold = Phi;
    Phi = Psi + at*phit';
    % end suivi Phi;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % HRHATRACK section [2]
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %% update of the eigenvalues/vectors
    if k==1 % first iteration
        [V,Lamb]=eig(Phi);
    end
    
    % update of the eigenvalues   
    iV = pinv(V);
    Lamb = Lamb*(1-muL)+muL*(iV*Phi*V);
    iLamb = diag(1./diag(Lamb));
    Lamb = diag(diag(Lamb));
    
    % update of the eigenvectors
    iLamb = inv(Lamb);
    V = (1-muV)*V + muV*(Phi*V*iLamb + Phi'*(V-Phi*V*iLamb)*iLamb');
    
    % normalization des colonnes de V
    for n = 1:Ns
        vn = V(:,n);
        vnorm = real(vn'*vn);
        V(:,n)= vn/sqrt(vnorm);
    end
    
    nu_i=angle(diag(Lamb))/2/pi;
    freqs(:,k) = nu_i;
    Vand = vdm(exp(j*2*pi*nu_i),P); 
    %cond(Vand)
    amps(:,k)=abs( pinv(Vand)*x(:) );%abs( Vand \ x(:) );
    
%    ampi = amps(:,k);
%    tab = [(1:Ns)' nu_i(:) ampi(:)];
%    sprintf('%d  %7.5f   %5.3e \n', tab')
%     pause
%     
  
   waitbar(k/Nit,hw)   
    
end

% for k=1:Ns
%     amps(k,:)=filter(1,[1 -beta],amps(k,:));
% end

close(hw);


