function signal = HR_synthesis(poles, amplitudes, signal_length);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% signal = HR_synthesis(poles, amplitudes, signal_length)
%
% Signal synthesis from the parameters of the EDS model
% O.Derrien, LMA Marseille & R.Badeau, LTCI Paris - 2009                        
%
% poles : vector of complex numbers representing damping and frequency
% amplitudes : vector of complex numbers representing magnitude and phase
%              magnitudes are normalized, i.e. represent the maximum
%              value of the exponential envelope
% signal_length : length of the signal to be synthesized, in samples
% signal : vector representing the audio signal
%
% Note that poles and amplitudes vectors must have the same size
% This function is the reciprocal of the HR_analysis function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[rows_p, columns_p] = size(poles);
[rows_a, columns_a] = size(amplitudes);
if (rows_p == rows_a) & (columns_p == columns_a)
   if rows_p > columns_p
       order = rows_p;
   else    
       order = columns_p;
       poles = poles.';
       amplitudes = amplitudes.';
   end
else
   disp('Error: poles and amplitudes vectors must have the same size');
   signal = 0;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Vandermonde matrix synthesis
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

order = length(poles);
V = zeros(signal_length, order);
for k = 1 : order
   z = poles(k);
   if abs(z) < 1,
       % positive damping case
       time = [0 : signal_length-1]';
       V(:,k) = z.^time;     
   else
       % negative damping case
       time = [-signal_length+1 : 0]';
       V(:,k) = z.^time;
   end
   V(:,k) = V(:,k) / norm(V(:,k));
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Signal synthesis
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

signal = V * amplitudes;
