%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Application script
%
% Signal modeling with exponentially damped sinusoids (EDS)
% Parameters estimation with the ESPRIT algorithm
% Analysis/synthesis in frequency subbands
%
% O.Derrien, LMA Marseille & R.Badeau, LTCI Paris - 2009
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

clear all
close all

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Analysis/Synthesis parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

N_max = 16384; % Maximum length of analysis windows (in samples)
P_os = 128; % Overlap at onsets (in samples)
P_stat = 2048; % Overlap in stationary parts (in samples), lower than N_max/2
              % Used when the maximum length of analysis windows is reached
SB = 8; % Number of subbands
taps = 256; % Filters length
order_sb = [4 4 4 4 4 4 4 4]; % Number of sines per subband
order_type = 1; % 1 = absolute value , 2 = relative value / samples in the subband

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Read the input signal
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Read the input signal
[input_signal, fs] = wavread('../../../audio/glockenSpielShort.wav');
L = length(input_signal);

% Length of the input signal is set to a multiple of SB
L = floor(L/SB)*SB;
input_signal = input_signal(1 : L);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Onset detection and segmentation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Onset detection
disp('Onset detection')
onset = onset_detection(input_signal, 0.01);
onset = onset(find(onset >= 2*P_os));

% Segmentation
bounds = [onset, L];
win_start = 1;
win_stop = [];
transition_type = 1; % 1 = onset (short transition), 0 = stationary part (long transition)
w = 1;
b = 1;
stop = 0;
while stop == 0
   win_stop_max = win_start(w) + N_max;
   if win_stop_max <= bounds(b)
       % maximum-length analysis window
       win_stop(w) = win_stop_max - 1;
       transition_type(w+1) = 0;
       win_start(w+1) = win_stop_max - P_stat;
   else
       % regular-length analysis window
       win_stop(w) = bounds(b);
       if b == length(bounds)
           stop = 1;
       else
           transition_type(w+1) = 1;
           win_start(w+1) = bounds(b) - P_os  + 1;
           b = b + 1;
       end
   end
   w = w + 1;
end
W = w - 1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computation of the filterbank
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Computing the filterbank')
[prototype_filter, analysis_filters, synthesis_filters] = analytic_PQMF(taps, SB);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Analysis loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Loop on time windows
disp('Analysis loop')
poles_sb = zeros(N_max/SB/4, SB, W);
amplitudes_sb = zeros(N_max/SB/4, SB, W);
components_sb = zeros(SB, W);
for w = 1 : W

   % Display the index of the current window
   fprintf('window %d/%d\n',w,W)

   % Current analysis window
   input_signal_win = input_signal(win_start(w) : win_stop(w));
   win_length = win_stop(w) - win_start(w) + 1;

   % Wide analysis window (for the compensation of filterbank group delay)
   input_signal_prev = zeros(taps/2, 1);
   input_signal_next = zeros(taps/2, 1);
   start1 = max(win_start(w)-taps/2, 1);
   start2 = max(win_start(w)-1, 1);
   if start2 > start1
       Ds = start2 - start1;
       input_signal_prev(end-Ds : end) = input_signal(start1 : start2);
   end
   stop1 = min(win_stop(w)+1, L);
   stop2 = min(win_stop(w)+taps/2, L);
   if stop2 > stop1
       Ds = stop2 - stop1 + 1;
       input_signal_next(1 : Ds) = input_signal(stop1 : stop2);
   end
   input_signal_win_wide = [input_signal_prev; input_signal_win; input_signal_next];

   % Loop on subbands
   for sb = 1 : SB

       % Analysis filtering
       input_signal_sb = filter(analysis_filters(:, sb), 1, input_signal_win_wide);

       % Downsampling
       input_signal_sb_ds = input_signal_sb(1 : SB : end);

       % Compensation of group delay
       input_signal_sb_ds = input_signal_sb_ds(taps/SB+1 : end);

       % Model order
       if order_type == 1
           % Absolute value
           order = order_sb(sb);
       elseif order_type == 2
           % Relative value
           order = round(order_sb(sb)*win_length/SB);
       else    
           order = 0
       end       
       order = min(order, round(win_length/SB/4));

       % Parameters estimation, without noise whitening
       [poles, amplitudes] = HR_analysis(input_signal_sb_ds, order, 0);

       % Storing parameters 
       poles_sb(1 : order, sb, w) = poles;
       amplitudes_sb(1 : order, sb, w) = amplitudes;
       components_sb(sb, w) = order;

   end    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Save the parameters in a file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

save('analysis_parameters.mat', 'poles_sb', 'amplitudes_sb', 'components_sb');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Synthesis loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Computing sine parts of synthesis weighting windows
sin_left_short = sin(pi/2*[0 : P_os-1]/P_os).^2;
sin_left_long = sin(pi/2*[0 : P_stat-1]/P_stat).^2;
sin_right_short = fliplr(sin_left_short);
sin_right_long = fliplr(sin_left_long);

% Loop on time windows
disp('Synthesis loop')
synth_signal = zeros(L, 1);
synth_signal_sb_prev2 = [];
synth_signal_sb_prev1 = [];
for w = 1 : W+1

   if w <= W

       % Display the index of the current window
       fprintf('window %d/%d\n',w,W)

       % Loop on subbands
       win_length = win_stop(w) - win_start(w) + 1;
       synth_signal_sb = zeros(win_length/SB, SB);
       for sb = 1 : SB

           % Recovering the parameters
           order = components_sb(sb, w);
           poles = poles_sb(1 : order, sb, w);
           amplitudes = amplitudes_sb(1 : order, sb, w);

           % Resynthesize subband signals
           if order ~= 0          
               synth_signal_sb(:, sb) = HR_synthesis(poles, amplitudes, win_length/SB);
           else
               synth_signal_sb(:, sb) = zeros(win_length/SB, 1);
           end           
       end
   end

   % Synthesis filterbank
   % Warning: Synthesizes the output signal in the previous time window   
   if w > 1

       % Wide synthesis window (for the compensation of filterbank group delay)
       if w == 2
           signal_temp_sb_prev2 = zeros(taps/SB/2, SB);
       else
           Ds = (win_stop(w-2) - win_start(w-1) + 1)/SB/2;
           signal_temp_sb_prev2 = synth_signal_sb_prev2(end-Ds-taps/SB/2+1 : end-Ds, :);
       end
       if w == W+1
           signal_temp_sb_cur = zeros(taps/SB/2, SB);
       else
           Ds = (win_stop(w-1) - win_start(w) + 1)/SB/2;
           signal_temp_sb_cur = synth_signal_sb(Ds+1 : Ds+taps/SB/2, :);
       end
       synth_signal_sb_prev1_wide = [signal_temp_sb_prev2; synth_signal_sb_prev1; signal_temp_sb_cur];

       % Loop on subbands
       win_length_prev1 = win_stop(w-1) - win_start(w-1) + 1;
       signal_temp_sb_prev1 = zeros(win_length_prev1, SB);
       for sb = 1 : SB

           % Oversampling
           signal_temp_interp = zeros(SB*size(synth_signal_sb_prev1_wide, 1), 1);
           signal_temp_interp(1 : SB : end) = synth_signal_sb_prev1_wide(:, sb);

           % Synthesis filtering
           signal_temp = filter(synthesis_filters(:, sb), 1, signal_temp_interp);

           % Compensation of group delay
           signal_temp_sb_prev1(:, sb) = signal_temp(taps : end-1);

       end

       % Combination of subband signals and coming back to a real signal
       synth_signal_win_prev1 = 2*SB*real(sum(signal_temp_sb_prev1, 2));

       % Computation of the synthesis weighting window
       if transition_type(w-1)
           if w == W+1
               synthesis_window = [sin_left_short, ones(1, win_length_prev1-2*P_os), sin_right_short];
           else
               if transition_type(w)
                   synthesis_window = [sin_left_short, ones(1, win_length_prev1-2*P_os), sin_right_short];
               else
                   synthesis_window = [sin_left_short, ones(1, win_length_prev1-P_stat-P_os), sin_right_long];
               end
           end
       else
           if w == W+1
               synthesis_window = [sin_left_long, ones(1, win_length_prev1-P_stat-P_os), sin_right_short];
           else
               if transition_type(w)
                   synthesis_window = [sin_left_long, ones(1, win_length_prev1-P_stat-P_os), sin_right_short];
               else
                   synthesis_window = [sin_left_long, ones(1, win_length_prev1-2*P_stat), sin_right_long];
               end
           end
       end

       % Overlap-add
       synth_signal(win_start(w-1) : win_stop(w-1)) = synth_signal(win_start(w-1) : win_stop(w-1)) + synth_signal_win_prev1 .* synthesis_window.';

   end

   % Refreshing buffers
   synth_signal_sb_prev2 = synth_signal_sb_prev1;
   synth_signal_sb_prev1 = synth_signal_sb;

end

% Visualizations
figure(1)
time = [1:L]/fs;
plot(time, input_signal)
hold on
plot(onset/fs, zeros(size(onset)),'r+')
hold off
xlabel('time(s)')
title('Input signal (blue) and onset positions (red +)')
drawnow
figure(2)
plot(time, input_signal)
hold on
plot(time, synth_signal, 'r')
hold off
xlabel('time(s)')
title('Time domain, input signal (blue) and synthetized signal (red)')
figure(3)
specgram(input_signal, 1024, fs)
title('Spectrogram, input signal') 
figure(4)
specgram(synth_signal,1024, fs)
title('Spectrogram, synthetized signal') 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Write the output signal in a wav file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

wavwrite(synth_signal, fs, '../../../output/highResolution_lma.wav')
